#Disclaimer
#This example is for illustrative purposes only and not for professional use.
#We accept no liability for any loss or damage as a result of the use of any
#form of this spreadsheet or its contents
#copyright 2009, Wthrich-Merz

#Data
#------------------------------------------------------------------------------
triangle1  <- read.csv2("C:/RData/1_GL.csv", header=F)
triangle1  <- as.matrix(triangle1)

triangle2  <- read.csv2("C:/RData/2_AL.csv", header=F)
triangle2  <- as.matrix(triangle2)

triangles  <- cbind(triangle1,triangle2)
triangles  <- array(triangles, dim=c(14,14,2))

v_i       <- read.csv2("C:/RData/v_i.csv", header=F)
v_i       <- as.matrix(v_i)

M1        <- triangles[,,1]/v_i[,1]
M2        <- triangles[,,2]/v_i[,2]

M         <- cbind(M1,M2)
M         <- array(M, dim=c(14,14,2))

v_array   <- array(data=NA, dim=c(2,2,14))
for (i in 1:14){
  v_array[,,i] <- v_i[i,]*diag(2)
}

# starting values m_j_(0)
#------------------------------------------------------------------------------

m         <- matrix(0, nrow=2, ncol=13)

for (j in 1:13){

  sum1  <- matrix(0, nrow=2, ncol=2)
  sum2  <- matrix(0, nrow=2, ncol=1)

  for (i in 1:(14-j)){

    sum1.add  <- v_array[,,i]
    sum1      <- sum1 + sum1.add

    sum2.add  <- v_array[,,i]%*%M[i,j+1,]
    sum2      <- sum2 + sum2.add
  }

  m[,j]  <- solve(sum1)%*%sum2

}


for(i in 1:3)
{

#Output
print("incremetal loss ratios")
print(m)
print("------------------------------------------------------------------------------------")

# covariance parameter sigma_j-1_(1)
#-------------------------------------------------------------------------------

sigma     <- array(data=NA, dim=c(2,2,13))

for (j in 1:13){

  sum1  <- matrix(0, nrow=2, ncol=2)

  for (i in 1:(14-j)){

    sum1.add  <- solve(v_array[,,i])^(0.5)%*%(triangles[i,(j+1),]-v_array[,,i]%*%m[,j])%*%t(triangles[i,(j+1),]-v_array[,,i]%*%m[,j])%*%solve(v_array[,,i])^(0.5)
    sum1      <- sum1 + sum1.add

  }

  sigma[,,j]    <- 1/(13-j)*sum1

}

sigma.matrix    <- matrix(0, nrow=2, ncol=13)

for (j in 1:12){

  sigma.matrix[,j]  <- sqrt(diag(sigma[,,j]))

}



# Extrapolation covariance parameters
#------------------------------------------------------------------------------

phi_11.11 <- min(sigma.matrix[1,11]^2/abs(sigma.matrix[1,10]),abs(sigma.matrix[1,10]))
phi_11.22 <- min(sigma.matrix[2,11]^2/abs(sigma.matrix[2,10]),abs(sigma.matrix[2,10]))
phi_11.21 <- min(sigma[2,1,11]^2/abs(sigma[2,1,10]),abs(sigma[2,1,10]))

sigma.11  <- c(phi_11.11,phi_11.22)
sigma.matrix[,12]=sigma.11

phi_12.11 <- min(sigma.matrix[1,12]^2/abs(sigma.matrix[1,11]),abs(sigma.matrix[1,11]))
phi_12.22 <- min(sigma.matrix[2,12]^2/abs(sigma.matrix[2,11]),abs(sigma.matrix[2,11]))
phi_12.21 <- min(phi_11.21^2/abs(sigma[2,1,11]),abs(sigma[2,1,11]))

sigma.12  <- c(phi_12.11,phi_12.22)
sigma.matrix[,13]=sigma.12

#Output
print("sigma")
print(sigma.matrix)
print("------------------------------------------------------------------------------------")


# Correlation coefficients
#------------------------------------------------------------------------------

rho.matrix      <- matrix(0, nrow=1, ncol=13)

for (j in 1:11){

  rho.matrix[,j]  <- sigma[2,1,j]/(sigma.matrix[1,j]*sigma.matrix[2,j])

}

rho.matrix[,12]   <- phi_11.21/(sigma.matrix[1,12]*sigma.matrix[2,12])
rho.matrix[,13]   <- phi_12.21/(sigma.matrix[1,13]*sigma.matrix[2,13])

#Output
print("rho")
print(rho.matrix)
print("------------------------------------------------------------------------------------")



# Sigma_(1)
#------------------------------------------------------------------------------

sigma_1 <- sigma

sigma_1[1,1,12]   <- phi_11.11^2
sigma_1[2,2,12]   <- phi_11.22^2
sigma_1[1,2,12]   <- sigma_1[2,1,12]  <- phi_11.11*phi_11.22*rho.matrix[,12]

sigma_1[1,1,13]   <- phi_12.11^2
sigma_1[2,2,13]   <- phi_12.22^2
sigma_1[1,2,13]   <- sigma_1[2,1,13]  <- phi_12.11*phi_12.22*rho.matrix[,13]

# Filling triangles and calculating reserves
#-------------------------------------------------------------------------------

triangles.filled  <- triangles

for (j in 1:13){

  for (i in (15-j):14){

    triangles.filled[i,(j+1),1] <- v_i[i,1]*m[1,j]
    triangles.filled[i,(j+1),2] <- v_i[i,2]*m[2,j]

  }

}

for (i in 1:14){
  triangles.filled[i,,1]  <- cumsum(triangles.filled[i,,1])
  triangles.filled[i,,2]  <- cumsum(triangles.filled[i,,2])
}

reserves  <- matrix(0,13,2)

for (i in 1:13){

  reserves[i,1] <- triangles.filled[(i+1),14,1]-triangles.filled[(i+1),(14-i),1]
  reserves[i,2] <- triangles.filled[(i+1),14,2]-triangles.filled[(i+1),(14-i),2]

}

reserve.agg <- sum(reserves)


#Output reserves
print("reserves single accident years")
print(reserves)
print("reserves aggregated accident years")
print(reserve.agg)
print("------------------------------------------------------------------------------------")


# CONDITIONAL MSEP
#------------------------------------------------------------------------------

# Single accident year
#------------------------------------------------------------------------------

msep.single   <- numeric(13)

# 1. Conditional Process Variance
#------------------------------------------------------------------------------

ones          <- rep(1,2)

cpv           <- numeric(13)

for (i in 2:14){

sum.sigma     <- matrix(0,2,2)
  
  for (j in (14-i+1):13){
  
  sum.sigma.add <- sigma_1[,,j]
  sum.sigma     <- sum.sigma + sum.sigma.add
  
  }
  
cpv[i-1] <- ones%*%v_array[,,i]^(0.5)%*%sum.sigma%*%v_array[,,i]^(0.5)%*%ones
sqrt.cpv.single<-sqrt(cpv)
}

#Output
print("square root conditional process variance single accident years")
print(sqrt.cpv.single)
print("------------------------------------------------------------------------------------")


# 2. Conditional Estimation Error
#------------------------------------------------------------------------------

cee           <- numeric(13)

for (i in 2:14){

  sum2     <- matrix(0,2,2)
  
  for (j in (14-i+1):13){
  
  
    sum1      <- matrix(0,2,2)
  
    for (l in 1:(14-j)){
  
        
      sum1.add    <- v_array[,,l]^(0.5)%*%solve(sigma_1[,,j])%*%v_array[,,l]^(0.5)
      sum1        <- sum1 + sum1.add
  
    }
  
    sum2.add  <- solve(sum1)
    sum2      <- sum2 + sum2.add
    
  }  
    
cee[i-1] <- ones%*%v_array[,,i]%*%sum2%*%v_array[,,i]%*%ones

}
sqrt.cee.single <- sqrt(cee)
msep.single          <- cpv + cee
sqrt.msep.single     <- sqrt(msep.single)


#Output
print("square root conditional estimation error single accident years")
print(sqrt.cee.single)
print("square root conditional MSEP single accident years")
print(sqrt.msep.single)
print("------------------------------------------------------------------------------------")


# Aggregated accident years
#------------------------------------------------------------------------------

summand1  <- sum(msep.single)
summand2  <- numeric(1)


for (i in 2:13){

   sum2     <- matrix(0,2,2) 
    
    for (j in (14-i+1):13){
    sum1     <- matrix(0,2,2)
        
      for (l in 1:(14-j)){
   
      sum1.add    <- v_array[,,l]^(0.5)%*%solve(sigma_1[,,j])%*%v_array[,,l]^(0.5)
      sum1        <- sum1 + sum1.add
      
      }
      
    sum2      <- sum2 + solve(sum1)
      
    } 
  
    for (k in (i+1):14){

    summand2.add  <- ones%*%v_array[,,i]%*%sum2%*%v_array[,,k]%*%ones
    summand2      <- summand2 + summand2.add
    
    }
    
}

sqrt.cpv.agg   <- sqrt(sum(cpv))
msep.agg    <- summand1 + 2*summand2
sqrt.msep.agg     <- sqrt(msep.agg)
sqrt.cee.agg   <- sqrt(sqrt.msep.agg^2-sum(cpv))

print("square root conditional process variance aggregated accident years")
print(sqrt.cpv.agg)
print("square root conditional estimation error aggregated accident years")
print(sqrt.cee.agg)
print("square root conditional MSEP aggregated accident years")
print(sqrt.msep.agg)
print("------------------------------------------------------------------------------------")



# Next iterations m_j_(1), m_j_(2)
#------------------------------------------------------------------------------

m_1         <- matrix(0, nrow=2, ncol=13)

for (j in 1:13){

  sum1  <- matrix(0, nrow=2, ncol=2)
  sum2  <- matrix(0, nrow=2, ncol=1)

  for (i in 1:(14-j)){

    sum1.add  <- v_array[,,i]^(0.5)%*%solve(sigma_1[,,j])%*%v_array[,,i]^(0.5)
    sum1      <- sum1 + sum1.add

    sum2.add  <- v_array[,,i]^(0.5)%*%solve(sigma_1[,,j])%*%v_array[,,i]^(0.5)%*%M[i,j+1,]
    sum2      <- sum2 + sum2.add
  }

  m_1[,j]  <- solve(sum1)%*%sum2

}

m <- m_1
print("------------------------------------------------------------------------------------")
print("------------------------------------------------------------------------------------")
}