#Disclaimer
#This example is for illustrative purposes only and not for professional use.
#We accept no liability for any loss or damage as a result of the use of any
#form of this spreadsheet or its contents
#copyright 2009, Wthrich-Merz

# Data
#-------------------------------------------------------------------------------

triangle1  <- read.csv2("C:/RData/1_GL.csv", header=F)
triangle1  <- as.matrix(triangle1)

triangle2  <- read.csv2("C:/RData/2_AL.csv", header=F)
triangle2  <- as.matrix(triangle2)

triangles  <- cbind(triangle1,triangle2)
x.triangles  <- array(triangles, dim=c(14,14,2))

c.triangles  <- x.triangles
for (i in 1:14){
  c.triangles[i,,1]  <- cumsum(x.triangles[i,,1])
  c.triangles[i,,2]  <- cumsum(x.triangles[i,,2])
}

v_i       <- read.csv2("C:/RData/v_i.csv", header=F)
v_i       <- as.matrix(v_i)

# Estimation of incremental loss ratios m_j and CL factors f_j 
#-------------------------------------------------------------------------------

# CL factors f_j
#-------------------------------------------------------------------------------

CLf         <- numeric(13)
CL.triangle  <- c.triangles[,,2]

for (j in 1:13){

  factor1 <- numeric(1)
  factor2 <- numeric(1)

  for (i in 1:(14-j)){

    factor1 <- factor1 + CL.triangle[i,j]
    factor2 <- factor2 + CL.triangle[i,(j+1)]

  }

  CLf[j] <- solve(factor1)%*%factor2

}

# incremental loss ratios m_j
#-------------------------------------------------------------------------------

vA          <- v_i[,1]
m           <- numeric(13)
ALR.triangle <- x.triangles[,,1]

for (j in 1:13){

  factor1 <- numeric(1)
  factor2 <- numeric(1)

  for (i in 1:(14-j)){

    factor1 <- factor1 + vA[i]
    factor2 <- factor2 + ALR.triangle[i,(j+1)]

  }

  m[j] <- solve(factor1)%*%factor2

}

# Filling triangles and calculating reserves
#-------------------------------------------------------------------------------

# CL triangle
#-------------------------------------------------------------------------------

CL.triangle.filled  <- CL.triangle

for (j in 1:13){

  for (i in (15-j):14){

    CL.triangle.filled[i,(j+1)] <- CL.triangle.filled[i,j]*CLf[j]

  }

}

CL.reserves  <- numeric(13)

for (i in 1:13){

  CL.reserves[i] <- CL.triangle.filled[(i+1),14]-CL.triangle.filled[(i+1),(14-i)]

}

CL.reserves.agg  <- sum(CL.reserves)
CL.Ultimate     <- CL.triangle.filled[,14]

# ALR triangle
#-------------------------------------------------------------------------------

ALR.triangle.filled   <- c.triangles[,,1]

for (j in 1:13){

  for (i in (15-j):14){

    ALR.triangle.filled[i,(j+1)] <- ALR.triangle.filled[i,j] + vA[i]*m[j]

  }

}

ALR.reserves  <- numeric(13)

for (i in 1:13){

  ALR.reserves[i] <- ALR.triangle.filled[(i+1),14]-ALR.triangle.filled[(i+1),(14-i)]

}

ALR.reserves.agg   <- sum(ALR.reserves)
ALR.Ultimate      <- ALR.triangle.filled[,14]


comb.reserves       <- ALR.reserves + CL.reserves
comb.reserves.agg <- sum(comb.reserves)

# parameter sigma_A
#-------------------------------------------------------------------------------

sigma.A     <- numeric(13)

for (j in 1:13){

  summe1  <- 0

  for (i in 1:(14-j)){

    summe1.add  <- vA[i]^(-1)*(x.triangles[i,(j+1),1]-vA[i]*m[j])^2
    summe1      <- summe1 + summe1.add

  }

  sigma.A[j]    <- 1/(13-j)*summe1

}

sigma.A[12] <- min(sigma.A[10],sigma.A[11]^2/sigma.A[10])
sigma.A[13] <- min(sigma.A[11],sigma.A[12]^2/sigma.A[11])

sqrt.sigma.A  <- sqrt(sigma.A)

# parameter sigma_C
#-------------------------------------------------------------------------------

# CL F
#-------------------------------------------------------------------------------

CLF   <- array(data=NA, dim=c(14,14,1))

for (j in 2:14){
  CLF[,j,] <- c.triangles[,j,2]/c.triangles[,j-1,2]
}

Q <- numeric(13)
w <- 1

w <- numeric(13)

for (j in 1:13){

  a1  <- matrix(0,1,1)
  a2  <- matrix(0,1,1)
  a3  <- matrix(0,1,1)

    for (l in 1:(14-j)){

    a1  <- a1 + sqrt(c.triangles[l,j,2])*sqrt(c.triangles[l,j,2])
    a2  <- a2 + c.triangles[l,j,2]
    a3  <- a3 + c.triangles[l,j,2]

    }

  w[j] <- a1^2/(a2*a3)

}

w <- 1

for (j in 1:13){

  Q[j]  <- 1/(13-j-1+w)

}

P <- numeric(13)

for (j in 1:13){

  summe <- numeric(1)

  for (i in 1:(14-j)){

    summe   <- summe + ((c.triangles[i,j,2])^0.5*(CLF[i,(j+1),1]-CLf[j]))^2

  }

  P[j]  <- summe

}

sigma.C <- Q*P

sigma.C[13] <- min(sigma.C[11],sigma.C[12]^2/sigma.C[11])


sqrt.sigma.C  <- sqrt(sigma.C)

# parameter sigma_CA
#-------------------------------------------------------------------------------

sigma.CA     <- numeric(13)

for (j in 1:13){

  summe1  <- 0

  for (i in 1:(14-j)){

    factor1 <- (c.triangles[i,j,2])^0.5*(CLF[i,(j+1),1]-CLf[j])
    factor2 <- (x.triangles[i,(j+1),1]-vA[i]*m[j])*vA[i]^(-0.5)
    summe1      <- summe1 + factor1*factor2

  }

  sigma.CA[j]    <- 1/(13-j)*summe1

}

sigma.CA[13] <- sigma.CA[12] <- abs(sigma.CA[11])

# rho
#-------------------------------------------------------------------------------

rho <- numeric(13)

for (j in 1:13){

  rho[j]  <- sigma.CA[j]/sqrt(sigma.A[j]*sigma.C[j])

}

rho[13] <- rho[12]  <- abs(rho[11])

################################################################################
# MSEP single accident years                                                   #
################################################################################

# ALR method
#-------------------------------------------------------------------------------

ALR.cpv <- numeric(13)

for (i in 1:13){

  summe <- 0

  for (j in (13-i+1):13){
  
  summe <- summe + sigma.A[j]
  
  }
  
  ALR.cpv[i]  <- vA[i+1]*summe
  sqrt.ALR.cpv.single <- sqrt(ALR.cpv)
}

ALR.cee <- numeric(13)

for (i in 2:14){

  summe2  <- 0

  for (j in (14-i+1):13){


    summe1  <- 0

    for (l in 1:(14-j)){

      summe1   <- summe1 + vA[l]*sigma.A[j]^(-1)

    }

    summe2      <- summe2 + summe1^(-1)

  }

ALR.cee[i-1] <- vA[i]^2*summe2
sqrt.ALR.cee.single <- sqrt(ALR.cee)

}

ALR.msep.single          <- ALR.cpv + ALR.cee
sqrt.ALR.msep.single     <- sqrt(ALR.msep.single)

# CL method
#-------------------------------------------------------------------------------

sigmaC_matrix <- matrix(0,13,13)

for (i in 1:13){

  for (j in 1:13){

  sigmaC_matrix[i,j]  <- CL.triangle.filled[(i+1),j]*sigma.C[j]

  }
}

helper      <- CLf[2:13]
helper      <- cumprod(helper[12:1])
helper      <- helper[12:1]

CLf.cumprod <- numeric(13)

for (j in 1:12){

  CLf.cumprod[j] <- helper[j]

}

CLf.cumprod[13] <- 1

CL.cpv   <- numeric(13)

for (i in 1:13){

  sum.1     <- 0

  for (l in (13-i+1):13){

    sum.1  <- sum.1 + CLf.cumprod[l]^2*sigmaC_matrix[i,l]

  }

  CL.cpv[i]  <- sum.1
  sqrt.CL.cpv.single <- sqrt(CL.cpv)
}



ww <- numeric(13)

for (j in 1:13){

  summe <- 0

  for (k in 1:(13-j+1)){

    summe <- summe + c.triangles[k,j,2]*sigma.C[j]^(-1)

  }

  ww[j] <- summe^(-1)

}

delta <- numeric(13)

    for (i in 1:13){

      summand1 <- 1
      part4    <- 1

      for (l in (14-i):13){

      part1   <- CLf[l]^2

      part2   <- ww[l]

      factor    <- part1 + part2

      summand1  <- summand1*factor

      part4     <- part4*part1

      }

      deltapart <- summand1 - part4

      delta[i]  <- deltapart

    }

CL.cee <- numeric(13)


for (i in 1:13){
  CL.cee[i]  <- (c.triangles[(i+1),(14-i),2])^2*delta[i]
}

sqrt.CL.cee.single <- sqrt(CL.cee)
CL.msep.single        <- CL.cpv + CL.cee
sqrt.CL.msep.single  <- sqrt(CL.msep.single)

# Cross product term
#-------------------------------------------------------------------------------

# Conditional cross process variance (ccpv)
#-------------------------------------------------------------------------------

sigmaCA_matrix <- matrix(0,13,13)

for (i in 1:13){

  for (j in 1:13){

  sigmaCA_matrix[i,j]  <- CL.triangle.filled[(i+1),j]^(0.5)*sigma.CA[j]*vA[i+1]^(0.5)
  
  }
}

ccpv   <- numeric(13)

for (i in 1:13){

  sum.1     <- 0

  for (l in (13-i+1):13){

    sum.1  <- sum.1 + CLf.cumprod[l]*sigmaCA_matrix[i,l]

  }

  ccpv[i]  <- sum.1

}

# Conditional cross estimation error (ccee)
#-------------------------------------------------------------------------------

W <- numeric(13)

for (j in 1:13){

  summe <- 0
  
  for (k in 1:(14-j)){
  
    summe <- summe + CL.triangle[k,j]*sigma.C[j]^(-1)
      
  }

  W[j]  <- summe^(-1)
  
}

U <- numeric(13)

for (j in 1:13){

  summe <- 0
  
  for (k in 1:(14-j)){
  
    summe <- summe + vA[k]*sigma.A[j]^(-1)
      
  }

  U[j]  <- summe^(-1)
  
}

T.j <- numeric(13)

for (j in 1:13){

  summe <- 0
  
  for (i in 1:(14-j)){
  
    summe <- summe + CL.triangle[i,j]^(0.5)*sigma.C[j]^(-1)*sigma.CA[j]*sigma.A[j]^(-1)*vA[i]^(0.5)
      
  }

  T.j[j]  <- W[j]*summe*U[j]
  
}

Psi.matrix  <- matrix(0,13,13)

for (k in 1:13){
  
  product <- prod(CLf[(14-k):13])

  for (i in 1:13){
  
  maxi     <- max((13-k+1),(13-i+1))
  summe <- 0
  
    for (j in maxi:13){
    
    summe <- summe + 1/CLf[j]*T.j[j]
    
    }
  
  Psi.matrix[k,i] <- product*summe
  
  }

}

ccee   <- numeric(13)

for (i in 1:13){

  ccee[i]  <- CL.triangle[(i+1),(14-i)]*Psi.matrix[i,i]*vA[i+1]

}

comb.cpv              <- sqrt(ALR.cpv + CL.cpv + 2*ccpv)
comb.cee              <- sqrt(ALR.cee + CL.cee + 2*ccee)
comb.msep.single      <- CL.msep.single + ALR.msep.single + 2*ccpv + 2*ccee
sqrt.comb.msep.single <- sqrt(comb.msep.single)

################################################################################
# MSEP aggregated accident years                                               #
################################################################################

# ALR method
#-------------------------------------------------------------------------------

summand1  <- sum(ALR.msep.single)
summand2  <- numeric(1)


for (i in 2:13){

   summe2     <- numeric(1)
    
    for (j in (14-i+1):13){
    summe1     <- numeric(1)
        
      for (l in 1:(14-j)){
   
      summe1  <- summe1 + vA[l]*(sigma.A[j])^(-1)
      
      }
      
    summe2  <- summe2 + (summe1)^(-1)
      
    } 
  
    for (k in (i+1):14){

    summand2  <- summand2 + vA[i]*summe2*vA[k]
        
    }
    
}

ALR.msep.agg      <- summand1 + 2*summand2
sqrt.ALR.msep.agg <- sqrt(ALR.msep.agg)

sqrt.ALR.cee.agg     <- sqrt(ALR.msep.agg-sum(ALR.cpv))
sqrt.ALR.cpv.agg     <- sqrt(sum(ALR.cpv))

# CL method
#-------------------------------------------------------------------------------

summand1  <- sum(CL.msep.single)
summand2  <- numeric(1)
product   <- numeric(1)

for (i in 2:13){

  teil.1  <- CL.triangle[i,(15-i)]*delta[(i-1)]
  
  for (k in (i+1):14){
  
  product  <- prod(CLf[(15-k):(14-i)])
  
  teil.2  <- CL.triangle[k,(15-k)]*product
    
  summand2  <- summand2 + teil.1%*%teil.2 
  
  }
   
}

CL.msep.agg       <- summand1 + 2*summand2
sqrt.CL.msep.agg  <- sqrt(CL.msep.agg)

sqrt.CL.cee.agg      <- sqrt(CL.msep.agg-sum(CL.cpv))
sqrt.CL.cpv.agg     <- sqrt(sum(CL.cpv))

# combined method
#-------------------------------------------------------------------------------

part1 <- sum(comb.msep.single)

part2 <- numeric(1)

for (i in 2:13){

  part2.1  <- CL.triangle[i,(15-i)]
  
  for (l in (i+1):14){
  
  part2.2  <- Psi.matrix[(l-1),(i-1)]*vA[l]
  
  part2   <- part2 + part2.1*part2.2 
  
  }
   
}

part3 <- numeric(1)

for (i in 2:13){

  part3.1  <- vA[i]
  
  for (l in (i+1):14){
  
  part3.2  <- CL.triangle[l,(15-l)]*Psi.matrix[(i-1),(l-1)]   
  
  part3   <- part3 + part3.1*part3.2 
  
  }
   
}

part4 <- numeric(1)

for (i in 2:13){

  part4.1  <- CL.triangle[i,(15-i)]*delta[(i-1)]
  
  for (l in (i+1):14){
  
  product <- prod(CLf[(15-l):(14-i)])
  part4.2  <- CL.triangle[l,(15-l)]*product
  
  part4   <- part4 + part4.1*part4.2 
  
  }
   
}

part5 <- numeric(1)

for (i in 2:13){

   summe2     <- numeric(1) 
    
    for (j in (14-i+1):13){
    summe1     <- numeric(1)
        
      for (r in 1:(14-j)){
   
      summe1    <- summe1 + vA[r]*(sigma.A[j])^(-1)
      
      }
      
    summe2      <- summe2 + (summe1)^(-1)
      
    } 
  
    for (l in (i+1):14){

    part5  <- part5 + vA[i]*summe2*vA[l]
    
    }
    
}

comb.msep.agg       <- part1 + 2*(part2 + part3 + part4 + part5)
sqrt.comb.msep.agg  <- sqrt(comb.msep.agg)

sqrt.comb.cpv.agg  <- sqrt(sum(comb.cpv^2))
sqrt.comb.cee.agg  <- sqrt(comb.msep.agg-sum(comb.cpv^2))



#Output
  print(CLf)
  #############################################################################################
  print(m)
  #############################################################################################
  print(rho)
  #############################################################################################
  print(CL.reserves)
  #############################################################################################
  print(CL.reserves.agg)
  #############################################################################################
  print(ALR.reserves)
  #############################################################################################
  print(ALR.reserves.agg)
  #############################################################################################
  print(comb.reserves)
  #############################################################################################
  print(comb.reserves.agg)
  #############################################################################################
  print(sqrt.ALR.cpv.single)
  #############################################################################################
  print(sqrt.ALR.cee.single)
  #############################################################################################
  print(sqrt.ALR.msep.single)
  #############################################################################################
  print(sqrt.CL.cpv.single)
  #############################################################################################
  print(sqrt.CL.cee.single)
  #############################################################################################
  print(sqrt.CL.msep.single)
  ############################################################################################# 
  print(comb.cpv)
  #############################################################################################
  print(comb.cee)
  #############################################################################################
  print(sqrt.comb.msep.single)
  #############################################################################################
  print(sqrt.ALR.cpv.agg)
  #############################################################################################
  print(sqrt.ALR.cee.agg)
  #############################################################################################
  print(sqrt.ALR.msep.agg)
  #############################################################################################
  print(sqrt.CL.cpv.agg)
  #############################################################################################
  print(sqrt.CL.cee.agg)
  #############################################################################################
  print(sqrt.CL.msep.agg)
  #############################################################################################
  print(sqrt.comb.cpv.agg)
  #############################################################################################
  print(sqrt.comb.cee.agg)
  #############################################################################################
  print(sqrt.comb.msep.agg)
  #############################################################################################