#Disclaimer
#This example is for illustrative purposes only and not for professional use.
#We accept no liability for any loss or damage as a result of the use of any
#form of this spreadsheet or its contents
#copyright 2009, Wthrich-Merz

#Data
#------------------------------------------------------------------------------
triangle1  <- read.csv2("C:/RData/1_GL.csv", header=F)
triangle1  <- as.matrix(triangle1)

triangle2  <- read.csv2("C:/RData/2_AL.csv", header=F)
triangle2  <- as.matrix(triangle2)

triangles  <- cbind(triangle1,triangle2)
triangles  <- array(triangles, dim=c(14,14,2))

for (i in 1:14){
  triangles[i,,1]  <- cumsum(triangles[i,,1])
  triangles[i,,2]  <- cumsum(triangles[i,,2])
}

# Chain-Ladder F
#-------------------------------------------------------------------------------

CLF   <- array(data=NA, dim=c(14,14,2))

for (j in 2:14){
  CLF[,j,] <- triangles[,j,]/triangles[,j-1,]
}

# starting values Chain-Ladder factor f
#------------------------------------------------------------------------------

CLf   <- matrix(0,2,13)

for (j in 1:13){

  faktor1 <- matrix(0,2,2)
  faktor2 <- matrix(0,2,1)

  for (i in 1:(14-j)){

    faktor1 <- faktor1 + diag(triangles[i,j,])
    faktor2 <- faktor2 + diag(triangles[i,j,])%*%CLF[i,(j+1),]

  }

  CLf[,j] <- solve(faktor1)%*%faktor2

}


for (k in 1:3){

#Output
print("CL factors")
print(CLf)
print("------------------------------------------------------------------------------------")

# sigma_j
#-------------------------------------------------------------------------------

sigma.sqr   <- matrix(0,2,13)

for (j in 1:13){

  summe <- matrix(0,2,1)

  for (i in 1:(14-j)){

    summe <- summe + (diag(CLF[i,(j+1),])-diag(CLf[,j]))^2%*%triangles[i,j,]

  }

  sigma.sqr[,j] <- 1/(13-j)*summe

}

sigma <- sqrt(sigma.sqr)


# sigma_12 (Extrapolation)
#-------------------------------------------------------------------------------
for (i in 1:2){
sigma[i,13]=min(sigma[i,12],sigma[i,11],sigma[i,12]^2/sigma[i,11])
}

#Output
print("sigma")
print(sigma)
print("------------------------------------------------------------------------------------")


# Estimation correlations rho_j
#-------------------------------------------------------------------------------

# Construction matrix Q
#-------------------------------------------------------------------------------

w_array   <- array(data=NA, dim=c(2,2,13))

for (j in 1:13){

  w.single  <- matrix(0,2,2)

  for (n in 1:2){
  for (m in 1:2){

  a1  <- matrix(0,1,1)
  a2  <- matrix(0,1,1)
  a3  <- matrix(0,1,1)

    for (l in 1:(14-j)){

    a1  <- a1 + sqrt(triangles[l,j,n])*sqrt(triangles[l,j,m])
    a2  <- a2 + triangles[l,j,n]
    a3  <- a3 + triangles[l,j,m]

    }

  w.single[n,m] <- a1^2/(a2*a3)

  }
  }

  w_array[,,j]  <- w.single

}

Q_array   <- array(data=NA, dim=c(2,2,13))

for (j in 1:13){
  for (n in 1:2){
  for (m in 1:2){

    Q_array[n,m,j]  <- 1/(14-j-2+w_array[n,m,j])
  }
  }
}

# Construction matrix P
#-------------------------------------------------------------------------------

P_array   <- array(data=NA, dim=c(2,2,13))

for (j in 1:13){

  summe <- numeric(1)

  for (l in 1:(14-j)){

    faktor1 <- solve(diag(sigma[,j]))%*%(diag(triangles[l,j,]))^0.5%*%(CLF[l,(j+1),]-CLf[,j])
    faktor2 <- t((CLF[l,(j+1),]-CLf[,j]))%*%(diag(triangles[l,j,]))^0.5%*%solve(diag(sigma[,j]))
    summe   <- summe + faktor1%*%faktor2

  }

  P_array[,,j]  <- summe

}

# Construction rho_j
#-------------------------------------------------------------------------------

cov_array <- array(data=NA, dim=c(2,2,13))

for (j in 1:13){

  cov_array[,,j]  <- P_array[,,j]*Q_array[,,j]

}

rho <- cov_array[2,1,]

# Extrapolation rho_11 und rho_12
#-------------------------------------------------------------------------------
cand1   <- abs(rho[11]*sigma[1,11]*sigma[2,11])
cand2   <- abs(rho[10]*sigma[1,10]*sigma[2,10])
cand3   <- cand1^2/cand2


phi.11  <- min(cand1, cand2, cand3)
rho[12] <- phi.11/(sigma[1,12]*sigma[2,12])

cand1   <- abs(rho[12]*sigma[1,12]*sigma[2,12])
cand2   <- abs(rho[11]*sigma[1,11]*sigma[2,11])
cand3   <- cand1^2/cand2


phi.12  <- min(cand1, cand2, cand3)
rho[13] <- phi.12/(sigma[1,13]*sigma[2,13])

cov_array[2,1,12] <- cov_array[1,2,12]  <- rho[12]
cov_array[2,1,13] <- cov_array[1,2,13]  <- rho[13]
cov_array[1,1,13] <- cov_array[2,2,13]  <- 1

#Output
print("rho")
print(rho)
print("------------------------------------------------------------------------------------")



# Filling triangles and calculating reserves
#-------------------------------------------------------------------------------

triangles.filled  <- triangles

for (j in 1:13){

  for (i in (15-j):14){

    triangles.filled[i,(j+1),1] <- triangles.filled[i,j,1]*CLf[1,j]
    triangles.filled[i,(j+1),2] <- triangles.filled[i,j,2]*CLf[2,j]

  }

}

reserves  <- matrix(0,13,2)

for (i in 1:13){

  reserves[i,1] <- triangles.filled[(i+1),14,1]-triangles.filled[(i+1),(14-i),1]
  reserves[i,2] <- triangles.filled[(i+1),14,2]-triangles.filled[(i+1),(14-i),2]

}

reserve.agg <- sum(reserves)

#Output
print("reserves single accident years")
print(reserves)
print("reserves aggregated accident years")
print(reserve.agg)
print("------------------------------------------------------------------------------------")


################################################################################
# MSEP Single accident year                                                    #
################################################################################

# Construction Sigma^C
#-------------------------------------------------------------------------------

sigma_array <- array(data=NA, dim=c(2,2,13))

for (j in 1:13){

  sigma_array[,,j]  <- diag(sigma[,j])%*%cov_array[,,j]%*%diag(sigma[,j])

}

sigmaC_array <- array(data=NA, dim=c(2,2,13,13))

for (i in 1:13){

  for (j in 1:13){

  sigmaC_array[,,j,i]  <- (diag(triangles.filled[(i+1),j,]))^0.5%*%sigma_array[,,j]%*%(diag(triangles.filled[(i+1),j,]))^0.5

  }
}

helper      <- CLf[,2:13]
helper[1,]  <- cumprod(helper[1,12:1])
helper[2,]  <- cumprod(helper[2,12:1])
helper      <- helper[,12:1]

CLf.cumprod <- array(data=NA, dim=c(2,2,13))

for (j in 1:12){

  CLf.cumprod[,,j] <- diag(helper[,j])

}

CLf.cumprod[,,13] <- diag(2)

# Estimation conditional process variance
#-------------------------------------------------------------------------------

ones  <- rep(1,2)
cpv   <- numeric(13)

for (i in 1:13){

  sum.1     <- matrix(0,2,2)

  for (l in (13-i+1):13){

    sum.1.add <- CLf.cumprod[,,l]%*%sigmaC_array[,,l,i]%*%CLf.cumprod[,,l]
    sum.1     <- sum.1 + sum.1.add

  }

  cpv[i]  <- ones%*%sum.1%*%ones
  sqrt.cpv.single <- sqrt(cpv)
}

#Output
print("square root conditional process variance single accident years")
print(sqrt.cpv.single)
print("------------------------------------------------------------------------------------")



################################################################################
# Construction matrix W (formula (8.36))                                       #
################################################################################

w_array <- array(data=NA, dim=c(2,2,13))

for (j in 1:13){

  summe <- matrix(0,2,2)

  for (k in 1:(13-j+1)){

    summe <- summe + diag(triangles[k,j,])^0.5%*%solve(sigma_array[,,j])%*%diag(triangles[k,j,])^0.5

  }

  w_array[,,j] <- solve(summe)

}


#
# delta.array.3
#-------------------
delta_array.3<- array(data=NA, dim=c(2,2,13))

for (n in 1:2){

  for (m in 1:2){

    for (i in 1:13){

      summand1 <- 1
      part4    <- 1

      for (l in (14-i):13){

      part1   <- CLf[n,l]*CLf[m,l]

      part2   <- w_array[n,m,l]

      faktor    <- part1 + part2

      summand1  <- summand1*faktor

      part4     <- part4*part1

      }

      deltapart <- summand1 - part4

      delta_array.3[n,m,i]  <- deltapart

    }

  }

}

cee <- numeric(13)

for (i in 1:13){
  cee[i]  <- ones%*%diag(triangles[(i+1),(14-i),])%*%delta_array.3[,,i]%*%diag(triangles[(i+1),(14-i),])%*%ones
}

sqrt.cee.single <- sqrt(cee)
msep.single <- cpv + cee
sqrt.msep.single     <- sqrt(msep.single)


#Output
print("square root conditional estimation error single accident years")
print(sqrt.cee.single)
print("square root conditional MSEP single accident years")
print(sqrt.msep.single)
print("------------------------------------------------------------------------------------")


################################################################################
# Conditional MSEP aggregated accident years                                   #
################################################################################

ones      <- rep(1,2)
summand1  <- sum(msep.single)
summand2  <- numeric(1)
produkt   <- numeric(2)

for (i in 2:13){

  part.1  <- ones%*%diag(triangles[i,(15-i),])%*%delta_array.3[,,(i-1)]

  for (k in (i+1):14){

  produkt[1]  <- prod(CLf[1,(15-k):(14-i)])
  produkt[2]  <- prod(CLf[2,(15-k):(14-i)])
  produkte    <- diag(produkt)

  part.2  <- diag(triangles[k,(15-k),])%*%produkte%*%ones

  summand2  <- summand2 + part.1%*%part.2

  }

}

sqrt.cpv.agg   <- sqrt(sum(cpv))
sqrt.msep.agg  <- sqrt(summand1 + 2*summand2)
sqrt.cee.agg   <- sqrt(sqrt.msep.agg^2-sum(cpv))

#Output
print("square root conditional process variance aggregated accident years")
print(sqrt.cpv.agg)
print("square root conditional estimation error aggregated accident years")
print(sqrt.cee.agg)
print("square root conditional MSEP aggregated accident years")
print(sqrt.msep.agg)
print("------------------------------------------------------------------------------------")


# Next iterations Chain-Ladder factor f
#------------------------------------------------------------------------------

CLf.1   <- matrix(0,2,13)

for (j in 1:13){

  faktor1 <- matrix(0,2,2)
  faktor2 <- matrix(0,2,1)

  for (i in 1:(14-j)){

    faktor1 <- faktor1 + diag(triangles[i,j,])^0.5%*%solve(sigma_array[,,j])%*%diag(triangles[i,j,])^0.5
    faktor2 <- faktor2 + diag(triangles[i,j,])^0.5%*%solve(sigma_array[,,j])%*%diag(triangles[i,j,])^0.5%*%CLF[i,(j+1),]

  }

  CLf.1[,j] <- solve(faktor1)%*%faktor2

}

CLf <- CLf.1
print("------------------------------------------------------------------------------------")
print("------------------------------------------------------------------------------------")
}